/*\
||| This file a part of Pike, and is copyright by Fredrik Hubinette
||| Pike is distributed as GPL (General Public License)
||| See the files COPYING and DISCLAIMER for more information.
\*/
#include "global.h"
RCSID("$Id: dbm.c,v 1.4 1998/05/15 19:06:49 grubba Exp $");
#include "dbm_machine.h"

/* Todo: make sure only one thread accesses the same ndbmmod */

#if defined(HAVE_NDBM_H) && defined(HAVE_DBM_OPEN)

#include "interpret.h"
#include "svalue.h"
#include "stralloc.h"
#include "array.h"
#include "object.h"
#include "pike_macros.h"

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#include "threads.h"
#include <ndbm.h>

/* Pair (key+data) block size */
#ifndef PBLKSIZ
#define PBLKSIZ	1024
#endif /* PBLKSIZ */

/* Data block size */
#ifndef DBLKSIZ
#define DBLKSIZ	4096
#endif /* DBLKSIZ */

#ifdef _REENTRANT
static MUTEX_T dbm_lock;
#endif  

struct dbm_glue
{
  DBM *dbf;
};

#define THIS ((struct dbm_glue *)(fp->current_storage))

static void do_free(void)
{
  if(THIS->dbf)
  {
    DBM *dbf;
    dbf=THIS->dbf;
    THIS->dbf=0;

    THREADS_ALLOW();
    mt_lock(& dbm_lock);
    dbm_close(dbf);
    mt_unlock(& dbm_lock);
    THREADS_DISALLOW();
  }
}

static int fixmods(char *mods)
{
  int mode;
  mode=0;
  while(1)
  {
    switch(*(mods++))
    {
    case 0:
      switch(mode & 15)
      {
      default: error("No mode given for dbm->open()\n"); 
      case 1|16:
      case 1: mode=O_RDONLY; break;
      case 3:
      case 3|16: mode=O_RDWR; break;
      case 7:
      case 7|16: mode=O_CREAT | O_RDWR; break;
      case 15:
      case 15|16: mode=O_CREAT | O_RDWR | O_TRUNC; break;
      }
      return mode;

    case 'r': case 'R': mode|=1;  break;
    case 'w': case 'W': mode|=3;  break;
    case 'c': case 'C': mode|=7;  break;
    case 't': case 'T': mode|=15; break;
    case 'f': case 'F': mode|=16; break;

    default:
      error("Bad mode flag in dbm->open.\n");
    }
  }
}

static void dbmmod_create(INT32 args)
{
  do_free();
  if(args)
  {
    DBM *tmp;
    struct pike_string *tmp2;
    int rwmode = O_CREAT | O_RDWR;

    if(sp[-args].type != T_STRING)
      error("Bad argument 1 to dbm->create()\n");

    if(args>1)
    {
      if(sp[1-args].type != T_STRING)
	error("Bad argument 2 to dbm->create()\n");

      rwmode=fixmods(sp[1-args].u.string->str);
    }

    tmp2=sp[-args].u.string;

    THREADS_ALLOW();
    mt_lock(& dbm_lock);
    tmp=dbm_open(tmp2->str, rwmode, 00666);
    mt_unlock(& dbm_lock);
    THREADS_DISALLOW();

    if (!tmp) {
      if (tmp2->len < 1024) {
	error("dbm->open(): Failed to open DBM database \"%s\"\n", tmp2->str);
      } else {
	error("dbm->open(): Failed to open DBM database\n");
      }
    }
    THIS->dbf=tmp;

    pop_n_elems(args);
  }
}

#define STRING_TO_DATUM(dat, st) dat.dptr=st->str,dat.dsize=st->len
#define DATUM_TO_STRING(dat) make_shared_binary_string(dat.dptr, dat.dsize)

static void dbmmod_fetch(INT32 args)
{
  DBM *dbf=THIS->dbf;
  datum key, ret;

  if(!args)
    error("Too few arguments to dbm->fetch()\n");

  if(sp[-args].type != T_STRING)
    error("Bad argument 1 to dbm->fetch()\n");

  if(!dbf)
    error("DBM database not open.\n");

  STRING_TO_DATUM(key, sp[-args].u.string);

  THREADS_ALLOW();
  mt_lock(& dbm_lock);
  ret=dbm_fetch(dbf, key);
  mt_unlock(& dbm_lock);
  THREADS_DISALLOW();

  pop_n_elems(args);
  if(ret.dptr)
  {
    push_string(DATUM_TO_STRING(ret));
  }else{
    push_int(0);
  }
}

static void dbmmod_delete(INT32 args)
{
  DBM *dbf = THIS->dbf;
  datum key;
  int ret;

  if(!args)
    error("Too few arguments to dbm->delete()\n");

  if(sp[-args].type != T_STRING)
    error("Bad argument 1 to dbm->delete()\n");

  if(!dbf)
    error("DBM database not open.\n");

  STRING_TO_DATUM(key, sp[-args].u.string);

  THREADS_ALLOW();
  mt_lock(& dbm_lock);
  ret=dbm_delete(dbf, key);
  mt_unlock(& dbm_lock);
  THREADS_DISALLOW();
  
  pop_n_elems(args);
  push_int(0);
}

static void dbmmod_firstkey(INT32 args)
{
  DBM *dbf=THIS->dbf;
  datum ret;
  pop_n_elems(args);

  if(!dbf) error("DBM database not open.\n");

  THREADS_ALLOW();
  mt_lock(& dbm_lock);
  ret=dbm_firstkey(dbf);
  mt_unlock(& dbm_lock);
  THREADS_DISALLOW();

  if(ret.dptr)
  {
    push_string(DATUM_TO_STRING(ret));
  }else{
    push_int(0);
  }
}

static void dbmmod_nextkey(INT32 args)
{
  DBM *dbf=THIS->dbf;
  datum ret;

  if(!dbf)
    error("DBM database not open.\n");

  THREADS_ALLOW();
  mt_lock(& dbm_lock);
  ret=dbm_nextkey(dbf);
  mt_unlock(& dbm_lock);
  THREADS_DISALLOW();

  pop_n_elems(args);
  if(ret.dptr)
  {
    push_string(DATUM_TO_STRING(ret));
  }else{
    push_int(0);
  }
}

static void dbmmod_store(INT32 args)
{
  DBM *dbf=THIS->dbf;
  datum key,data;
  int ret;
  if(args<2)
    error("Too few arguments to dbm->store()\n");

  if(sp[-args].type != T_STRING)
    error("Bad argument 1 to dbm->store()\n");

  if(sp[1-args].type != T_STRING)
    error("Bad argument 2 to dbm->store()\n");

  if(!dbf)
    error("DBM database not open.\n");

  if (sp[-args].u.string->len + sp[1-args].u.string->len >= PBLKSIZ-2) {
    error("String-pair too large to dbm->store()\n");
  }

  STRING_TO_DATUM(key, sp[-args].u.string);
  STRING_TO_DATUM(data, sp[1-args].u.string);

  THREADS_ALLOW();
  mt_lock(& dbm_lock);
  ret=dbm_store(dbf, key, data, DBM_REPLACE);
  mt_unlock(& dbm_lock);
  THREADS_DISALLOW();

  if(ret == -1)
    error("DBM database not open for writing.\n");

  pop_n_elems(args);
  push_int(ret == 0);
}

static void dbmmod_close(INT32 args)
{
  pop_n_elems(args);

  do_free();
  push_int(0);
}

static void init_dbm_glue(struct object *o)
{
  THIS->dbf=0;
}

static void exit_dbm_glue(struct object *o)
{
  do_free();
}

#endif /* HAVE_NDBM_H && HAVE_DBM_OPEN */

void pike_module_exit(void) {}

void pike_module_init(void)
{
#if defined(HAVE_NDBM_H) && defined(HAVE_DBM_OPEN)
  start_new_program();
  add_storage(sizeof(struct dbm_glue));
  
  add_function("create",dbmmod_create,"function(void|string,void|string:void)",0);

  add_function("close",dbmmod_close,"function(:void)",0);
  add_function("store",dbmmod_store,"function(string,string:int)",0);
  add_function("`[]=",dbmmod_store,"function(string,string:int)",0);
  add_function("fetch",dbmmod_fetch,"function(string:string)",0);
  add_function("`[]",dbmmod_fetch,"function(string:string)",0);
  add_function("delete",dbmmod_delete,"function(string:int)",0);
  add_function("firstkey",dbmmod_firstkey,"function(:string)",0);
  add_function("nextkey",dbmmod_nextkey,"function(string|void:string)",0);

  set_init_callback(init_dbm_glue);
  set_exit_callback(exit_dbm_glue);
  end_class("dbm",0);
#endif /* HAVE_NDBM_H && HAVE_DBM_OPEN */
}

